--[[
Plugin that displays bars with given duration.
There is 3 anchors that the bars move between depending on the duration left.
They can either animate between anchors in both movement and size or
just move there instantly without animation (controlled in the options)
]]

----------------------------------
-- Create plugin

local RW = LibStub("AceAddon-3.0"):GetAddon("Raid Watch")
local plugin = RW:Plugin("Bars")
if not plugin then return end

----------------------------------
-- Locals

local L = LibStub("AceLocale-3.0"):GetLocale("RW2-Plugin-Bars")
local LB = LibStub("AceLocale-3.0"):GetLocale("RW2-Plugin-Base")
local LSM = LibStub("LibSharedMedia-3.0")

local Utils = RW.Utils
local db

local anchorOrder = {"Short", "Normal", "Long"}

----------------------------------
-- Bar prototype
-- Start, Stop, SetText, SetTime, SetElapsed
-- Update, ApplySettings, SetPosition, AnimateSize
-- Animate, GetOrder

local barProt = {}


----------------------------------
-- Default options

local defaults = {
	profile = {
		fadeTime = 0.5,
		fadeOutTime = 2,
		flashTime = 10,
		flashSpeed = 2,
		flashColor = {1, 0, 0},
		PluginEnabled = true,
		Enabled = {
			Long = true,
			Normal = true,
		},
		Master = {
			UseGlobal = true,
			GrowUp = true
		},
		Animate = true,
		AnimSpeed = 2,
		FillUp = true,
		Anchors = {
			["**"] = {
				UseGlobal = false,
				UseMaster = true,
				Fill = true,
				GrowUp = true,
				Enabled = true,
				x = 0,
				y = 0,
			},
			Short = {
				Size = {Width = 290, Height = 46, Spacing = 0},
				Time = 12,
			},
			Long = {
				Size = {Width = 280, Height = 32, Spacing = 0},
				Time = 1000000,
			},
			Normal = {
				Size = {Width = 280, Height = 32, Spacing = 0},
				Time = 70,
			},
		}
	}
}

-- TODO: Check what the Master actually contains after this
-- Fetch the default settings (fonts, textures and so on)
-- Defined in Common.lua
for i, setting in pairs(Utils.defaults.profile) do
	defaults.profile.Master[i] = setting
end

-- Backdrop table used for all the bars
local inset = defaults.profile.Master.BorderSize > 2 and defaults.profile.Master.Insets or 0
local backdrop = {
	bgFile = LSM:Fetch("background", defaults.profile.Master.BGTexture), 
	edgeFile = LSM:Fetch("border", defaults.profile.Master.Border), 
	tile = false,
	edgeSize = defaults.profile.Master.BorderSize,
	insets = { left = inset, right = inset, top = inset, bottom = inset },
}

--------------------------
-- Init

function plugin:OnRegister()
	self.db = RW.db:RegisterNamespace(self:GetName(), defaults)
	db = self.db.profile
	
	self.db.RegisterCallback(self, "OnProfileChanged", "Reset")
	self.db.RegisterCallback(self, "OnProfileCopied", "Reset")
	self.db.RegisterCallback(self, "OnProfileReset", "Reset")
end

function plugin:PluginEnable()
	if not db.PluginEnabled then
		self:Disable()
		return
	end
	self:RegisterMessage("RW_ConfigMode")
	self:RegisterMessage("RW_DefaultsChanged")
	self:RegisterMessage("RW_StartBar")
	self:RegisterMessage("RW_CancelBar")
	
	self.freeBarObjects = self.freeBarObjects or {}
	self.runningBars = self.runningBars or {}
	
	self:SendMessage("RW_RegisterCommListener", self, "StartCommBar")
end

function plugin:PluginDisable()
	if self.runningBars then
		for bar in pairs(self.runningBars) do
			bar:Stop()
		end
		
		wipe(self.freeBarObjects)
		wipe(self.runningBars)
	end	
end



local function onUpdate(self, elapsed)
	for bar in pairs(plugin.runningBars) do
		bar:Update(elapsed)
	end
end

function plugin:EnableUpdate(enable)
	if not self.updateFrame then self.updateFrame = CreateFrame("Frame", "RW_Bars_UpdateFrame") end
	if enable then
		self.updateFrame:SetScript("OnUpdate", onUpdate)
	else
		self.updateFrame:SetScript("OnUpdate", nil)
	end
end

----------------------------------
-- Anchors

local function dragStart(self)
	self:StartMoving()
end

local function dragStop(self)
	self:StopMovingOrSizing()
	local scale = self:GetEffectiveScale() / UIParent:GetEffectiveScale()
	
	local x, y = self:GetCenter()
	x, y = x * scale, y * scale
	
	x = x - GetScreenWidth()/2
	y = y - GetScreenHeight()/2
	
	local aDb = db.Anchors[self.anchorName]
	
	aDb.x = x
	aDb.y = y
	
	plugin:SendMessage("RW_GUIValuesChanged")
end

local function createAnchor(name)
	local db = db.Anchors[name] or db.Master
	plugin.anchors = plugin.anchors or {}
	
	local frame = Utils:CreateAnchor(name, UIParent, name.." bars")
	
	frame:ClearAllPoints()
	local scale = frame:GetEffectiveScale() / UIParent:GetEffectiveScale()
	frame:ClearAllPoints()
	frame:SetPoint("CENTER", UIParent, "CENTER", db.x / scale, db.y / scale)		
	frame:SetWidth(db.Size.Width)
		
	frame:SetScript("OnDragStart", dragStart)
	frame:SetScript("OnDragStop", dragStop)
	frame.bars = {}
	frame.anchorName = name
	plugin.anchors[name] = frame
end

local function createAnchors()
	for i, setting in pairs(plugin.db.profile.Anchors) do
		createAnchor(i)
	end
end

local function rightClick(self, button)
	if button == "RightButton" then
		self.obj:Stop()
	elseif button == "LeftButton" then
		local party = RW:GetPlugin("Party")
		local pType = ((party:GetPartyType() == "raid" and "RAID") or (party:GetPartyType() == "party" and "PARTY") or (party:GetPartyType() == "battleground" and "BATTLEGROUND")) or nil
		if pType then
			SendChatMessage(self.obj.orgText.." ending in [" ..self.obj.bar.timeText:GetText().."]", pType)
		end
	end
end

function plugin:Reset()
	db = self.db.profile
	if self.anchors then
		for anchorName, anchor in pairs(self.anchors) do
			local aDb = db.Anchors[anchorName]
			anchor:ClearAllPoints()	
			local scale = anchor:GetEffectiveScale() / UIParent:GetEffectiveScale()
			anchor:ClearAllPoints()
			anchor:SetPoint("CENTER", UIParent, "CENTER", aDb.x / scale, aDb.y / scale)	
			
			self:UpdateAnchor(anchor) 
			self.anchors[anchorName]:SetWidth(aDb.Size.Width)
		end
	end
	
	if self.runningBars  then
		for bar in pairs(self.runningBars) do
			bar.anchorSettings = db.Anchors[bar.anchor.anchorName]
		end
	end
end

-------------------------------
-- Bar prototype functions

function barProt:Start()
	plugin.runningBars[self] = true
	plugin:EnableUpdate(true)
end

function barProt:Stop()
	plugin.runningBars[self] = nil
	plugin.freeBarObjects[self] = true
	self.anchor.bars[self] = nil
	plugin:UpdateAnchor(self.anchor)
	self.bar:SetAlpha(0)
	self.bar:Hide()
	self.id = nil
	self.text = nil
	self.timeLeft = nil
	self.totalTime = nil
	self.icon = nil
	self.fadeTime = nil
	self.anchor = nil
	self.flashDir = nil
	self.flashTime = nil
	self.bar:UnregisterAllEvents()
	if not next(plugin.runningBars) then plugin:EnableUpdate(false) end
end

function barProt:SetText(text)
	self.bar.text:SetText(text or self.text)
end

function barProt:SetTime(time)
	self.totalTime = time
	self.timeLeft = time
end

function barProt:SetElapsed(time)
	self.timeLeft = self.totalTime - time
end

function barProt:Update(elapsed)
	self.timeLeft = self.timeLeft - elapsed
	local db = plugin.db.profile
	-- Bar has ended
	if self.timeLeft < 0 then
		-- Bar just ended
		if not self.fadeOutTime then
			self.fadeOutTime = db.fadeOutTime
			plugin:SendMessage("RW_BarFinished", self.id, self.text, self.icon)
		-- Fade out ended
		elseif self.fadeOutTime < 0.1 then
			self:Stop()
			self.fadeOutTime = nil
			return
		-- Update fadeout
		else
			self.fadeOutTime = self.fadeOutTime - elapsed
			self.bar:SetAlpha(self.fadeOutTime / db.fadeOutTime)
		end
	-- Running bar
	else
		-- Just started
		if self.fadeTime then
			local newAlpha = (self.totalTime - self.timeLeft) / self.fadeTime
			if newAlpha >= 0.95 then
				self.bar:SetAlpha(1)
				self.fadeTime = nil
			else
				self.bar:SetAlpha(newAlpha)
			end
		elseif self.timeLeft < db.flashTime then
			self.flashTime = (self.flashTime or 0)
			
			local perc
			if self.flashTime <= db.flashSpeed / 2 then
				perc = self.flashTime / (db.flashSpeed / 2)
			else
				perc = (db.flashSpeed - self.flashTime) / (db.flashSpeed / 2)
			end
					
			if self.flashDir then
				if self.flashTime > db.flashSpeed then
					self.flashDir = false
				end
			else
				if self.flashTime < 0 then
					self.flashDir = true
				end
			end	
			
			local r, g, b
			if self.color then
				r, g, b = unpack(self.color)
			else
				r, g, b = unpack(db.Master.BarColor)
			end
			self.bar.bar:SetStatusBarColor(Utils:ColorGradientSimple(perc, r, g, b, unpack(db.flashColor)))
			
			if self.flashDir then
				self.flashTime = self.flashTime + elapsed
			else
				self.flashTime = self.flashTime - elapsed
			end
		end
		-- Update text and statusbar
		self.bar.timeText:SetFormattedText("%s", Utils:TimeToStringShort(self.timeLeft))
		if db.FillUp then
			self.bar.bar:SetValue(1 - self.timeLeft/self.totalTime)
		else
			self.bar.bar:SetValue(self.timeLeft/self.totalTime)
		end
	end
	
	-- Find anchor (change anchor if needed, moveElapsed and changeAnchor will be set if anchor changed
	-- or if another bar on that anchor ended or moved)
	plugin:FindAnchor(self)
	
	-- Move to new anchor
	if self.moveElapsed then
		-- Animation enabled
		if db.Animate then 
			self.moveElapsed = self.moveElapsed - elapsed
			self:Animate(elapsed)
			-- Reached end of animation
			if self.moveElapsed < 0 then
				self.moveElapsed = nil
				self:SetPosition()
			end
		-- No animation
		else
			self:SetPosition()
			self.moveElapsed = nil
		end
	end
	
	-- Set Size to the new anchors settings
	if self.changeAnchor then 
		-- Animate enabled
		if db.Animate then 
			self.changeAnchor = self.changeAnchor - elapsed
			self:AnimateSize(elapsed) 
			-- Reached end of animation
			if self.changeAnchor < 0 then 
				self.changeAnchor = nil 
			end
		-- No animation
		else
			self.bar:SetWidth(self.anchorSettings.Size.Width)
			self.bar:SetHeight(self.anchorSettings.Size.Height)
			self.changeAnchor = nil
		end
	end
end

-- Apply visual settings
function barProt:ApplySettings()
	local db = plugin.db.profile.Master
	local bar = self.bar
	
	-- Set the insets to 0 if the bar have no border
	local inset
	if db.BorderSize < 2 then
		inset = 0
	else
		inset = db.Insets
	end
	
	-- Update backdrop
	backdrop.insets.left, backdrop.insets.right = inset, inset
	backdrop.insets.top, backdrop.insets.bottom = inset, inset	
	backdrop.bgFile = LSM:Fetch("background", db.BGTexture)
	backdrop.edgeFile = LSM:Fetch("border", db.Border)
	backdrop.edgeSize = db.BorderSize
	
	-- Colors
	bar:SetBackdrop(backdrop)
	bar:SetBackdropColor(unpack(db.BGColor))
	bar:SetBackdropBorderColor(unpack(db.BorderColor))		
	
	-- Statusbar
	bar.bar:SetStatusBarTexture(LSM:Fetch("statusbar", db.BarTexture))
	
	-- Texts
	bar.timeText:SetFont(LSM:Fetch("font", db.Font), db.FontSize, db.FontOutline)	
	bar.text:SetFont(LSM:Fetch("font", db.Font), db.FontSize, db.FontOutline)
	bar.text:SetJustifyH(db.FontAlign)
	bar.text:SetTextColor(unpack(db.FontColor))
	
	-- Icon
	bar.icon:SetTexture(self.icon)
	bar.icon:SetTexCoord(0.08, 0.92, 0.08, 0.92)
	
	-- Use custom color if existing, else use default
	if self.color then
		bar.bar:SetStatusBarColor(unpack(self.color))
	else
		bar.bar:SetStatusBarColor(unpack(db.BarColor))
	end
end

-- Set position (used at the end of animation or when no animation is used)
function barProt:SetPosition()
	local aDB = self.anchorSettings
	local anchor = self.anchor
	
	-- Using order to determin the bar offset
	-- Anchoring bars to eachother is not working, all kinds of dependancy errors on SetPoint
	local i = self:GetOrder()
	if aDB.GrowUp then
		self.bar:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT", 0 , aDB.Size.Height * i + aDB.Size.Spacing * i)
	else
		self.bar:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT", 0, -anchor:GetHeight() - aDB.Size.Height - aDB.Size.Height * i - aDB.Size.Spacing * i)
	end
	self.bar:SetWidth(aDB.Size.Width)
	self.bar:SetHeight(aDB.Size.Height)
	
	local margin
	if db.Master.BorderSize < 2 or db.Master.Border == "None" then 
		margin = 0 else margin = db.Master.BorderMargin 
	end
	if self.icon then
		self.bar.bar:SetWidth(self.bar:GetWidth() - self.bar:GetHeight() - margin)
	else
		self.bar.bar:SetWidth(self.bar:GetWidth() - margin * 2)
	end
end

-- Animate the bar when moving
function barProt:Animate(elapsed)
	local aDB = self.anchorSettings
	local anchor = self.anchor
	
	-- Get Start and Target coords (Y value needs more calculation since it depends on
	-- if bars grow up or down, spacing and order.
	local startX, startY = self.bar:GetLeft(), self.bar:GetBottom()
	local targetX, targetY = anchor:GetLeft(), 0
	
	local order = self:GetOrder()
	if aDB.GrowUp then
		targetY = anchor:GetTop() + aDB.Size.Height * order + aDB.Size.Spacing * order
	else
		targetY = anchor:GetTop() - aDB.Size.Height * order - aDB.Size.Spacing * order - anchor:GetHeight() - aDB.Size.Height
	end
	
	-- Linearpolate the movement 
	local newX = Utils:Lerp(startX, targetX, elapsed * db.AnimSpeed)
	local newY = Utils:Lerp(startY, targetY, elapsed * db.AnimSpeed)
	local diffx = -(targetX - newX)
	local diffy = -(targetY - newY)
	
	if aDB.GrowUp then
		self.bar:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT", diffx, diffy + aDB.Size.Height * order + aDB.Size.Spacing * order)
	else
		self.bar:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT", diffx, diffy - anchor:GetHeight() - aDB.Size.Height - aDB.Size.Height * order - aDB.Size.Spacing * order)
	end
end

-- Animate the size of the bar when moving to new anchor
function barProt:AnimateSize(elapsed)
	local aDB = self.anchorSettings
	local master = plugin.db.profile.Master
	local anchor = self.anchor
	local newWidth = Utils:Lerp(self.bar:GetWidth(), aDB.Size.Width, elapsed * db.AnimSpeed)
	local newHeight = Utils:Lerp(self.bar:GetHeight(), aDB.Size.Height, elapsed * db.AnimSpeed)
	self.bar:SetWidth(newWidth)
	self.bar:SetHeight(newHeight)
	
	local margin
	if master.BorderSize < 2 or master.Border == "None" then margin = 0 else margin = master.BorderMargin end
	if self.icon then
		self.bar.bar:SetWidth(self.bar:GetWidth() - self.bar:GetHeight() - margin)
	else
		self.bar.bar:SetWidth(self.bar:GetWidth() - margin * 2)
	end
end

-- Get the order for the bar. That is its time remaining relative to the other bars on the same anchor.
-- Have to put
do
	local function barSort(a, b)
		return a.timeLeft < b.timeLeft
	end
	
	local sortt = {}
	function barProt:GetOrder()
		wipe(sortt)
		for bar in pairs(self.anchor.bars) do
			tinsert(sortt, bar)
		end
		table.sort(sortt, barSort)
		for i, bar in pairs(sortt) do
			if bar == self then return i-1 end
		end
	end
end

-- Get a running bar, nil if not found
function plugin:GetBar(id)
	for bar in self:BarIterator() do
		if bar.id == id then return bar end
	end
end

-- Find anchor wich the bar should belong to
function plugin:FindAnchor(bar)
	local bestAnchor
	for i, anchor in pairs(db.Anchors) do
		-- We have at least found one anchor
		if bestAnchor then
			-- If the best anchor so far has less max time then the current time left of the bar
			if db.Anchors[bestAnchor].Time < bar.timeLeft then
				-- If the new anchors time is more then currect anchors time then its better then the previous one
				if anchor.Time >= db.Anchors[bestAnchor].Time then
					bestAnchor = i
				end
			-- If the new anchors time is less then the best so far and the time is more then the time left of the bar => better anchor found
			elseif anchor.Time < db.Anchors[bestAnchor].Time and anchor.Time >= bar.timeLeft then
				bestAnchor = i
			end
		-- No anchor found yet, then just take the current one
		elseif anchor.Enabled then
			bestAnchor = i
		end
	end
	
	local oldAnchor = bar.anchor
	local newAnchor = self.anchors[bestAnchor]
	
	-- If the current anchor is not the same as the new one, then the bar need to change anchor
	if oldAnchor and oldAnchor ~= newAnchor then
		oldAnchor.bars[bar] = nil
		self:UpdateAnchor(oldAnchor)
		
		newAnchor.bars[bar] = true
		bar.anchor = newAnchor
		bar.anchorSettings = db.Anchors[bestAnchor]		
		bar.totalTime = bar.anchorSettings.Time
		bar.changeAnchor = 4		
		self:UpdateAnchor(newAnchor)
		
	-- If the bar did not have an anchor yet
	elseif not bar.anchor then
		bar.anchorSettings = db.Anchors[bestAnchor]
		bar.anchor = newAnchor
		newAnchor.bars[bar] = true
		bar.changeAnchor = 4
		self:UpdateAnchor(newAnchor)
	end
end

------------------------------
-- Function for all bars

-- Update the anchor by setting all its bars to moving
function plugin:UpdateAnchor(anchor)
	anchor:SetWidth(db.Anchors[anchor.anchorName].Size.Width)
	for bar in pairs(anchor.bars) do
		bar.moveElapsed = 4
		bar.changeAnchor = 4
	end
end

function plugin:MoveAnchor(anchor)
	anchor:ClearAllPoints()
	local aDB = db.Anchors[anchor.anchorName]
	local s = anchor:GetEffectiveScale()
	anchor:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", aDb.x / s, aDb.y / s)
end

-- Itterator for all running bars
function plugin:BarIterator()
	return pairs(self.runningBars)
end


------------------------------
-- Message Handling 

--- Start a new bar
-- @param id Unique ident for the bar, used to control the bar (Stop)
-- @param duration The total time the bar should last for
-- @param text The text show on the bar
-- @param icon The icon, if not supplied no icon will be shown
-- @param color Table with the color (should be in indexed format e.g. {1, 0, 1, 1})
function plugin:RW_StartBar(message, id, duration, text, icon, color)
	if not duration or duration <= 0 then return end
	if not self.anchors then createAnchors() end
	local newBar = self:GetBar(id)
	if newBar then newBar:Stop() end
	newBar = next(self.freeBarObjects)
	if not newBar then
		newBar = setmetatable({
			bar = Utils:CreateBar(rightClick)
		}, {__index = barProt})
	else
		self.freeBarObjects[newBar] = nil
	end
	
	newBar.bar:RegisterEvent("MODIFIER_STATE_CHANGED")
	
	newBar.id = id	
	newBar.orgText = text
	newBar.text = Utils:ColorNames(text)
	newBar.timeLeft = duration
	newBar.totalTime = duration
	newBar.icon = icon
	newBar.color = color
	newBar.fadeTime = self.db.profile.fadeTime
	newBar.bar:Show()
	newBar.bar.obj = newBar
	
	self:FindAnchor(newBar)
	
	newBar:ApplySettings()
	newBar:SetPosition()
	newBar:SetText()
	newBar:Start()
end

function plugin:StartCommBar(sender, dist, id, duration, message)
	self:SendMessage("RW_StartBar", id, duration, message)
end

--- Cancel a running bar
-- @param id The id of the bar, same as the one supplied when starting a bar
function plugin:RW_CancelBar(message, id)
	local bar = self:GetBar(id)
	if bar then bar:Stop() end
end

function plugin:RW_DefaultsChanged(message, option, value)
	self.db.profile.Master[option] = value
	for bar in pairs(self.runningBars) do
		bar:ApplySettings()
	end
end

-- Config mode
do
	local counter = 0
	local function randomBars()
		local spellName, rank, icon = Utils:GetRandomSpell()
		plugin:RW_StartBar("2", counter, math.random(10, 100), spellName..(rank ~= "" and (" ("..rank..")") or ""), icon)
		counter = counter + 1
		if counter > 100 then counter = 0 end
	end
	
	local schedule
	function plugin:RW_ConfigMode(message, start)
		if start then
			if not self.anchors then createAnchors() end
			for i, anchor in pairs(self.anchors) do
				anchor:Show()
			end
			schedule = self:ScheduleRepeatingTimer(randomBars, 8)
			randomBars()
		else
			for i, anchor in pairs(self.anchors) do
				anchor:Hide()
			end
			self:CancelTimer(schedule)
			for i = 0, 100 do
				self:RW_CancelBar("RW_CancelBar", i)
			end
		end
	end
end

----------------------------------
-- Option table generation

function plugin:GetOptions()
	
	return LB.PLUGINS, LB.ALERT_PLUGINS, L.NAME, self.GetGUI, [[Interface\Icons\INV_Sigil_Thorim]]
end	

local function changeCallback()
	plugin:Reset()
end

local function changeCallback2(widget, event, value)
	local anchor = widget:GetUserData("anchor")
	plugin:UpdateAnchor(anchor)
end

function plugin:GetGUI()
	local self = self or plugin
	local UIF = LibStub("AceGUI-Factory-1.0"):GetFactory("RW")
	local AceGUI = LibStub("AceGUI-3.0")
	
	local group, label, slider, check, igroup, igroup2, button, color, select
	
	group = AceGUI:Create("ScrollFrame")
	group:SetLayout("Flow")
	
	label = UIF:MainTitleLabel(L.NAME)
	group:AddChild(label)
	
	igroup = UIF:InlineGroup("")
	group:AddChild(igroup)	
	
	label = UIF:NormalText(L.DESC)
	label:SetImage([[Interface\Icons\INV_Sigil_Thorim]])
	igroup:AddChild(label)
	
	igroup = UIF:InlineGroup("")
	group:AddChild(igroup)	
	
	check = UIF:CheckBox(LB.ENABLED, db, "PluginEnabled", function() 
		if not db.PluginEnabled then
			self:Disable()
		else
			self:Enable()
		end
	end)
	igroup:AddChild(check)
	
	check = UIF:CheckBox("Fill up", db, "FillUp", changeCallback)
	check:SetDescription("Toggle of bars should fill up or empty out")
	igroup:AddChild(check)
	
	igroup2 = UIF:InlineGroup2(L.ANIMATION)
	igroup:AddChild(igroup2)
	
	check = UIF:CheckBox(L.ANIMATE_TRANSITIONS, db, "Animate", changeCallback)
	check:SetDescription(L.ANIMATE_TRANSITIONS_DESC)
	check:SetRelativeWidth(0.5)
	igroup2:AddChild(check)
	
	slider = UIF:Slider(L.ANIMATE_SPEED, db, "AnimSpeed", 0.1, 5, 0.1, changeCallback)
	igroup2:AddChild(slider)
			
	igroup2 = UIF:InlineGroup2(L.BAR_DURATION)
	igroup:AddChild(igroup2)
	
	label = UIF:NormalText(L.BAR_DURATION_DESC)
	igroup2:AddChild(label)
	
	slider = UIF:Slider(L.SHORT, db.Anchors["Short"], "Time", 3, 100, 1, function(widget, event, value)
		if value >= db.Anchors["Normal"].Time then value = db.Anchors["Normal"].Time - 1 end
		db.Anchors["Short"].Time = value
		plugin:UpdateAnchor(plugin.anchors["Short"])
	end)
	igroup2:AddChild(slider)
	
	slider = UIF:Slider(L.NORMAL, db.Anchors["Normal"], "Time", 10, 200, 1, function(widget, event, value)
		if value < db.Anchors["Short"].Time then db.Anchors["Short"].Time = value - 1 end
		db.Anchors["Normal"].Time = value
		plugin:UpdateAnchor(plugin.anchors["Normal"])
	end)
	igroup2:AddChild(slider)
	
	igroup2 = UIF:InlineGroup2(L.FLASH)
	igroup:AddChild(igroup2)
	
	slider = UIF:Slider(L.FLASH_TIME, db, "flashTime", 1, 100, 1)
	igroup2:AddChild(slider)
	
	slider = UIF:Slider(L.FLASH_SPEED, db, "flashSpeed", 0.1, 4, 0.1)
	igroup2:AddChild(slider)
	
	color = UIF:ColorSelect(L.FLASH_COLOR, db, "flashColor")
	igroup2:AddChild(color)
	
	igroup2 = UIF:InlineGroup2(LB.SIZE)
	igroup:AddChild(igroup2)
	
	for _, i in pairs(anchorOrder) do
		label = UIF:NormalText(i..":")
		label:SetRelativeWidth(0.15)
		igroup2:AddChild(label)

		slider = UIF:Slider(LB.WIDTH, db.Anchors[i].Size, "Width", 10, 600, 1, changeCallback2)
		slider:SetUserData("anchor", plugin.anchors and plugin.anchors[i] or nil)
		slider:SetRelativeWidth(0.28)
		igroup2:AddChild(slider)
		
		slider = UIF:Slider(LB.HEIGHT, db.Anchors[i].Size, "Height", 10, 600, 1, changeCallback2)
		slider:SetUserData("anchor",  plugin.anchors and plugin.anchors[i] or nil)
		slider:SetRelativeWidth(0.28)
		igroup2:AddChild(slider)
		
		check = UIF:CheckBox(LB.GROW_UP, db.Anchors[i], "GrowUp", changeCallback2)
		check:SetRelativeWidth(0.28)
		check:SetUserData("anchor",  plugin.anchors and plugin.anchors[i] or nil)
		igroup2:AddChild(check)
	end
	
	igroup2 = UIF:InlineGroup2(LB.POSITION)
	igroup:AddChild(igroup2)
	
	for _, i in pairs(anchorOrder) do
		label = UIF:NormalText(i..":")
		label:SetRelativeWidth(0.15)
		igroup2:AddChild(label)

		slider = UIF:Slider("X", db.Anchors[i], "x",
			-floor(GetScreenWidth() * UIParent:GetEffectiveScale()), 
			floor(GetScreenWidth() * UIParent:GetEffectiveScale()),
			1, changeCallback)
		slider:SetRelativeWidth(0.42)
		igroup2:AddChild(slider)
		
		slider = UIF:Slider("Y", db.Anchors[i], "y",
			-floor(GetScreenHeight() * UIParent:GetEffectiveScale()), 
			floor(GetScreenHeight() * UIParent:GetEffectiveScale()),
			1, changeCallback)
		slider:SetRelativeWidth(0.42)
		igroup2:AddChild(slider)
	end
	
	
	button = UIF:Button(LB.TEST, function()
		local spellName, rank, icon = Utils:GetRandomSpell()
		--plugin:SendMessage("RW_StartBar", math.random(), math.random(10, 100), spellName..(rank ~= "" and (" ("..rank..")") or ""), icon)
		plugin:SendMessage("RW_StartBar", math.random(), 10, spellName..(rank ~= "" and (" ("..rank..")") or ""), icon)
		--local spellName, rank, icon = Utils:GetRandomSpell()
		--plugin:SendMessage("RW_StartBar", math.random(), 10, spellName..(rank ~= "" and (" ("..rank..")") or ""), icon)
		local spellName, rank, icon = Utils:GetRandomSpell()
		plugin:SendMessage("RW_StartBar", math.random(), 15, spellName..(rank ~= "" and (" ("..rank..")") or ""), icon)
		local spellName, rank, icon = Utils:GetRandomSpell()
		plugin:SendMessage("RW_StartBar", math.random(), 15, spellName..(rank ~= "" and (" ("..rank..")") or ""), icon)
	end)
	group:AddChild(button)
			
	button = UIF:Button(LB.RESET, function() 
		plugin.db:ResetProfile() 
	end)
	button:SetUserData("TriggerGlobalUpdate", true)
	button:SetUserData("confirm", true)
	group:AddChild(button)
	
	group:FixScroll()
	
	return group
end